
      subroutine rkfs (f,neqn, y,t,tout, relerr,abserr, iflag,
     1                 yp , h  , f1   , f2   , f3,
     2                 f4 , f5 , savre, savae,
     3                 nfe, kop, init , jflag, kflag)
c     DOUBLE PRECISION (real*8)
c     multi-step solver for Fehlberg 4-th/5-th order Runge-Kutta O.D.E.

c       program status:                     6 May 1990

c       J. G. Dworski

c       written by:  H. A. Watts and L. F. Shampine
c                    Sandia Laboratories
c                    Albuquerque, New Mexico

c       /rkfs/ is designed for non-stiff and mildly stiff O.D.E.,
c       assuming that derivative evaluations are not too expensive.
c       It exercises a COMPLETELY DATA ADAPTIVE stepping strategy
c      (the user is not even permitted to suggest an initial step size).
c       It is controlled merely by the specified error criterion and
c       the desired output break-points. Unlike a predictor-corrector,
c       /rkfs is a "marching" algorithm that practically cannot be
c       stalled by "adverse" data. It reports the level of "adversity",
c       and is automatically parametrized to forge on, if so instructed.

c       H a n d l i n g :
c       -----------------

c       /rkfs (the multi-step solver) is called by /rkf45 (the
c       solution "interfacing" routine), and /rkf45 is called by
c       Mainline.

c       Mainline calls on /rkf45, having formulated and declared
c       "external" the  s i n g l e  subroutine f(t,y,yp) the name of
c       which (illustrated by f) is arbitrary, but whose arguments
c       are  e x a c t l y  (t,y,yp).  /t/ is the (scalar) time,
c       /y/ is the vector of y-values, /yp/ is the vector of dy/dt
c       values, both of size /neqn/. f(t,y,yp) describes the system
c       of /neqn/ simultaneous ODEs.
c       All other pertinent info that f needs, to describe the ODE
c       system must be conveyed to it by means of labeled commons.

c       /rkfs calls on /fehl (the one-step solver).


c       A b s t r a c t

c       /rkfs/ integrates a system of /neqn/ 1-st order O.D.E.s
c       having form :
c               dy(i)/dt  =  f(t,y(1),y(2),...,y(neqn)) ,
c       where : the y(i) are given at time /t/.

c       Typically,
c       /rkfs/ is called by /rkf45/ to integrate from /t/ to /tout/
c       (see below for the non-trivial looping "syntax"),  but it can
c       be alternatively used as a "one-step integrator" to advance the
c       solution a single step in the direction of /tout/.  On return,
c       the parameters in the calling sequence are "set" for continu-
c       ing the integration.  The user has only to call /rkf45/ again
c       - defining a new value for /tout/ - .

c       Thus /rkf45/ is an "interfacing" routine :
c       it calls the "solution" subroutine /rkfs/. In turn, /rkfs/
c       calls the subroutine /fehl/ which computes an approximate
c       solution over one step.

c       M e t h o d

c       Runge-Kutta-Fehlberg (4,5) algorithm described in :
c               E. Fehlberg : "Low-order classical Runge-Kutta
c                              formulas with stepsize control",
c                              NASA Tech.Rep. R-315.
c       (available at the U.of W. Engr.Libr. : 629.1771 UN32TR)

c       the performance of /rkf45/ is evaluated in :
c       L. F. Shampine, H. A. Watts and S. Davenport :
c               "Solving non-stiff ordinary differential equations,
c                the state of the art",
c                Sandia Laboratories Report  SAND-75-0182, March 1975
c       (available at the U.of W. Engr.Libr. as microfishe, 71 pages),
c       and in the article (same authors and title) in :
c       SIAM Review, Vol. 18, No.3, July 1976, pp 376 - 411
c       (available at the U.of W. Math.Res.Libr. : 510.5 SIA v.18)

c       P a r a m e t e r s

c       f       -- subroutine f(t,y,yp), to evaluate derivatives
c                  yp(i) = dy(i)/dt
c       neqn    -- number of equations to be integrated
c       y(*)    -- solution vector at time /t/
c       t       -- (time), the independent variable of integration
c       tout    -- output time (point) at which solution is desired
c       relerr  -- relative error (tolerance) for "local error" test,
c                  for UW-APL/vax MAKE relerr.ge.1.0E-07
c                  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
c       abserr  -- absolute error (tolerance) for "local error" test
c               note: the code requires at each step, and for each
c                     component of the "local error" and "solution"
c                     vector, that :
c                 (abs("local error")) .le. (relerr*abs(y) + abserr)
c       iflag   -- "status of integration" indicator

c       w o r k   arguments (for internal use) :
c       six   a r r a y s  of minimal size /neqn/ :
c       /yp, f1, f2, f3, f4, f5/
c       three real    variables :
c       /h, savre, savae/
c       five  integer variables :
c       /nfe, kop, init, jflag, kflag/
c       note -- do not alter any work argument between calls,
c               but some of them are of possible diagnostic use:
c               yp  -- vector of solution derivatives at (time) t,
c(work(neqn+1)) h   -- step-size to be used in next integration step,
c(iwork(1)_     nfe -- counter of how many derivative evaluations were
c                      made up to this point (cumulant for all
c                      calls to /rkfs/, not exceeding 25000)

c   REMARKS (on parameter statement below)
c  (a)  /remin/ is minimum acceptable value of /relerr/, it depends
c       on machine implementation; attempts to obtain higher
c       accuracy with this subroutine are expensive and sometimes
c       unsuccessful :
c       MACHINE EPSILON ON 32 bit/word machines is APPROXIMATELY 1.5E-07,
c       and in doub. prec. 64 bit/word it is APPROXIMATELY 1.0E-16 .
c  (b)  the run expense is "bounded" by restricting the number
c       of allowed function derivative evaluations in any one
c       cumulative sequence of calls to /rkfs/ ...
c       by the flag /maxnfe=25000. It corresponds to about 4000
c       elementary computational intervals .
c  (c)  if the "output" (i.e., "tapping") of /rkfs/ is requested so
c       often that the "increment" of the independent argument is
c       twice finer than the step size which /rkfs/ deems sufficient
c       for the specified accuracy, then a counter (at STA 080+1)
c       monitors such occurrences and makes /rkfs/ return w/ IFLAG=7
c       when the count exceeds /koutput=2000 .

c     ------------------------------------------------------------
c     changes:
c       changed "double precision" to "real*8"  (craig friesen 7/8/92)
c     ------------------------------------------------------------

c       ===============================================================
        implicit real*8  (a-h,o-z)

        external  f

        parameter(remin=1.0d-16,maxnfe=25000,koutput=2000)

        real*8 y(neqn),t,tout,relerr,abserr,h
        real*8 yp(neqn),f1(neqn),f2(neqn),f3(neqn)
        real*8 f4(neqn),f5(neqn)
c-        real    savre,savae
c-        real    a,ae,dt,ee,eeoet,esttol,et,hmin,remin,rer,s,scale
c-        real    tol,toln,u26,epsp1,eps,ypk, amax1,amin1
c-        integer neqn,iflag,nfe,kop,init,jflag,kflag
c-        integer k,maxnfe,mflag

        logical hfaild,output

        common/NOTHING/  dt,u26

c       s t a r t   o f   e x e c u t a b l e s

c       check input parameters
c       ----------------------
        if (neqn.lt.1)                      goto 010
        if (relerr.le.0.d0 .or. abserr.lt.0.d0) goto 010
        mflag   =  iabs(iflag)
        if (mflag.eq.0   .or. mflag.gt.8)   goto 010
        if (mflag.ne.1)                     goto 020

c       compute "machine epsilon"
c       -------------------------
        eps   = 1.d0
  005   eps   = eps/2.d0
        epsp1 = eps + 1.d0
        if (epsp1.gt.1.d0)   goto 005
        u26   = 26.d0*eps
        rer   = 2.d0*eps + remin
c       MACHINE EPSILON ON 32 bit machines is APPROXIMATELY  1.2E-07
c       and in doub.prec. (64 bit/word) it is APPROXIMATELY  1.0D-16
        goto 050

c       invalid input = return
c       ----------------------
  010   iflag = 8
        r e t u r n
c       ~~~~~~~~~~~

c       determine conditions for continuation of integration
c       ====================================================
  020   if (t.eq.tout       .and. kflag.ne.3) goto 010
        if (                      mflag.ne.2) goto 025

c       reset-to-normal return : (kflag=+3 or kflag=+4)
c       .............................................
        if (init.eq.0       .or.  kflag.eq.3) goto 045
        if (                      kflag.eq.4) goto 040

c       abnormal    return : (iflag=+5 or iflag=+6)
c       ..............................................
        if (abserr.eq.0.d0    .and. kflag.eq.5) goto 030
        if (relerr.le.savre .and. abserr.le.savae
     1                      .and. kflag.eq.6) goto 030
        goto 050

c       reset-to-normal return : (iflag=+3 or iflag=+4)
c       .............................................
  025   if (                      iflag.eq.3) goto 045
        if (                      iflag.eq.4) goto 040
        if (abserr.gt.0.d0    .and. iflag.eq.5) goto 045

c       since user did not take remedial action to iflag=5 or 6 :
c       ........................................................
  030   s t o p
c       ~~~~~~~
c       ~~~~~~~

c       reset function derivative evaluation counter
c       ...............................................
  040   nfe   =   0
        if (                      mflag.eq.2) goto 050

c       reset flag value to that from the previous call
c       ...............................................
  045   iflag = jflag
        if (kflag.eq.3)   mflag = iabs(iflag)

c       normal return (iflag=+2 or -2); save input flag and
c       set continuation flag for input check on next step
c       ....................................................
  050   jflag = iflag
        kflag =   0

c       save /relerr/ and /abserr/ for input check on next step
c       .......................................................
        savre = relerr
        savae = abserr

c       set internal minimum relative error to at least three times
c       machine epsilon :
c       .................
c--  see above:  rer   = 2.*eps + remin
        if (relerr.ge.rer)   goto 055

c       relative error tolerance was too small and is increased
c       to an internally estimated value
c       ................................
        relerr = rer
        iflag  = 3
        kflag  = 3
        r e t u r n
c       ~~~~~~~~~~~

c       /dt/ is distance between current and target position
  055   dt     = tout - t

        if (mflag .eq. 1)  goto 060
        if (init  .eq. 0)  goto 065
        goto 080


c       i n i t i a l i z a t i o n
c       ===========================
c       (a) set "initialization completion" indicator : /init/
c       (b) set "too many output points"    indicator : /kop/
c       (c)      evaluate  initial derivatives
c       (d) set "derivative evaluation"     counter   : /nfe/
c       (e)      estimate starting  step-size
  060   init   = 0
        kop    = 0
        a      = t

c       .d0#.d0#.d0#.d0#.d0.d0
        call f( a,y,yp )
c       .d0#.d0#.d0#.d0#.d0.d0
        nfe    = 1
        if (t.ne.tout)     goto 065
        iflag  = 2
        r e t u r n
c       ~~~~~~~~~~~
  065   init   = 1
        h      = abs(dt)
        toln   = 0.d0
        do 070 k = 1,neqn
          tol  = relerr*abs(y(k)) + abserr
          if(tol.le.0.d0)    goto 070
          toln = tol
          ypk  = abs(yp(k))
          if (ypk*h**5 .gt. tol)  h = (tol/ypk)**0.2d0
  070   continue
        if (toln .le. 0.d0)  h = 0.d0
        h      = dmax1( h, u26*dmax1( abs(t),abs(dt) ) )
        jflag  = isign( 2,iflag )


c       set step-size, properly oriented: from /t/ towards /tout/
  080   h      = sign( h,dt )

c       determine whether /rkfs/'s efficiency is degraded by
c       request for too many output points
c       ...................................
        if (abs(h) .ge. 2.d0*abs(dt)) kop  = kop + 1
        if (kop    .ne.   koutput ) goto 085

c       when /kop/ reaches the value /koutput/ it becomes
c       obvious that too many otputs were requested (relative to
c       a step-size that is commensurate to the requested accuracy)
        kop    =  0
        iflag  =  7
        r e t u r n
c       ~~~~~~~~~~~

c      NOTE: occasional baseless "floating point exceptions" tend to occur
c            in the statement 085 when run on the PYRAMID ("slew").
c            I suspect this has to do with PYRAMID's emphasis on use
c            of registers, and heuristically preempt the problem by
c            placing /dt,u26 into a "NOTHING" common - which will prevent
c            the PYRAMID from storing these variables in registers :
  085   if (abs(dt) .gt. u26*abs(t))  then
          goto 095
        endif

c       if sufficiently close to targeted output point, then
c       extrapolate to it and return
c       ............................
        do 090 k = 1, neqn
  090   y(k)   = y(k) + dt*yp(k)
        a      = tout
c       .d0#.d0#.d0#.d0#.d0.d0
        call f( a,y,yp )
c       .d0#.d0#.d0#.d0#.d0.d0
        nfe    = nfe  + 1
        goto 300

c       initialize output point indicator
c       .................................
  095   output = .false.

c       scale error tolerances to avoid possible underflow in
c       the "error tolerance function" /esttol/; cf. STA 250+
c       .....................................................
        scale  = 2.d0/relerr
        ae     = scale*abserr

c       step-by-step   i n t e g r a t i o n
c       ====================================

  100   hfaild  = .false.

c       set the minimum permitted step-size
        hmin    = u26*abs(t)

c       adjust step-size if necessary to hit the target output point
c       (look ahead two steps to avoid sharp changes in step-size
c        and thus lessen impact of output points on the code quality)
c       .............................................................
        dt      = tout - t
        if (abs(dt) .ge. 2.d0*abs(h))   goto 200
        if (abs(dt) .gt.    abs(h))   goto 150

c       the next step (if successful) will complete the integration
c       to the targeted output point
c       ............................
        output  = .true.
        h       =  dt
        goto 200

  150   h       =  dt/2.d0


c       ..............................................................

c       single-step "core  i n t e g r a t o r "
c       ========================================

c       features:
c       (a) tolerances have been scaled to avoid underflow in
c           computing the "error tolerance function" /esttol/
c           (cf. STA 095+)
c       (b) relative error is referenced to average of absolute
c           magnitudes of the solution at the beginning and end of
c           of a step
c       (c) the error estimate formula has been "grouped" to
c           control loss of significance
c       (d) step-size /h/ is not permitted to become smaller than
c           26 times the roundoff error in the argument /t/, in
c           order to distinguish clearly between the various
c           arguments
c       (e) reasonable limits on changes in step-size are enforced
c           to "smoothen" the step-size selection process and avoid
c           "chatter" in problems that exhibit discontinuities
c       (f) after the code estimates a step-size that will likely
c           succeed, it uses only 9/10 of it, to afford a safety
c           margin against a possible "step failure"
c       (g) after a "step failure" the code does not permit the
c           step-size to increase for the  n e x t  attempted step
c           (this makes the code more efficient on problems that
c            exhibit discontinuities, as well as more effective
c            in general - inasmuch as extra caution seems warranted
c            when local extrapolation is used)

c       test number of function derivative evaluations
c       (if O.K. then try to advance the integration from t to t+h)
c       ...........................................................
  200   if (nfe .le. maxnfe)  goto 220

c       more steps than prudently allowed; pause and inspect :
c       ................
        iflag   =  4
        kflag   =  4
        r e t u r n
c       ~~~~~~~~~~~

c       advance the approximate solution over one step of length /h/
c       ============================================================

c       .d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0.d0
  220   call fehl( f,neqn,y,t,h,yp,f1,f2,f3,f4,f5,f1 )
c       .d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0#.d0.d0
        nfe     =  nfe + 5

c       make local error estimates and compare to allowable tolerances,
c       note "referencing" relative error (see (b) above)
c       ...............................................................

        eeoet   =  0.d0
        do 250 k= 1, neqn
          et    = abs(y(k)) + abs(f1(k)) + ae
        if (et    .gt.  0.d0)  goto 240
c       t r o u b l e, if solution vanishes.and.you specified abserr=0
        iflag   =  5
        r e t u r n
c       ~~~~~~~~~~~
  240   ee      = abs( (-2090.d0*yp(k) + (21970.d0*f3(k)
     1                - 15048.d0*f4(k)) )
     2                +(22528.d0*f2(k) -  27360.d0*f5(k) ) )
        eeoet   = dmax1( eeoet, ee/et )
  250   continue

        esttol  = abs(h) * eeoet * scale / 752400.d0
  255   if (esttol .le. 1.d0)  goto 260

c       unsuccessful step ( "step failure" ) :
c       try again with reduced step size (subject to (f) and (g) above)
c       ...............................................................
        hfaild  = .true.
        output  = .false.
        s       =  0.1d0
        if (esttol .lt. 59049.d0)  s = 0.9d0 / esttol**0.2d0
        h       =  s*h
        if (abs(h).gt.hmin)  goto 200

c       requested error unattainable even at smallest allowed step-size
c       ...............................................................
        iflag   =  6
        kflag   =  6
        r e t u r n
c       ~~~~~~~~~~~

c       SUCCESSFUL STEP :
C                  store solution at t+h and evaluate derivatives there
c                  ....................................................
  260   continue
        t       =  t + h

        do 270 k= 1, neqn
          y(k)  = f1(k)
  270   continue
        a       =  t
c       .d0#.d0#.d0#.d0#.d0.d0
        call f( a,y,yp )
c       .d0#.d0#.d0#.d0#.d0.d0
        nfe     = nfe  + 1

c       next step-size (increase limited to factor 5.and.(g) above)
c       ...........................................................
        s       =  5.d0
        if (esttol .gt. 1.889568D-04)  s = 0.9d0 / esttol**0.2d0
        if (hfaild                  )  s = dmin1( s,1.d0 )
        h       = sign( dmax1(s*abs(h),hmin), h )

c       END of single-step "core  i n t e g r a t o r"
c       ==============================================

c       ..............................................................


c       SHOULD WE TAKE ANOTHER STEP ??
c       ===========================

        if (output      )  goto 300
        if (iflag .gt. 0)  goto 100

c       integration successfully completed
c       ==================================

c       in "single-step" mode :
c       .....................
        iflag   = -2
        r e t u r n
c       ~~~~~~~~~~~

c       in "interval"    mode :
c       .....................
  300   t       = tout
        iflag   = +2
        r e t u r n
c       ~~~~~~~~~~~

        end
