      subroutine lapatch(x,y,z,dzx,dzy,mode)
c     DOUBLE PRECISION
c     membrane (Laplacian=0) patch interpolator for z
c     at longitude x and latitude y

c     subr/lapatch      status    : 6 May 1993

cf    cf comments made by craig friesen, to reduce stdout traffic

c     J. G. Dworski

c     H a n d l i n g  :
c     ==================
c     The "z-field" (assumed "positive", except as noted by "s p e c i a l"
c     below)   is fitted (on a unit square, with local co-ordinates
c     (xi,eta) [0,1] ) with a membrane having the 4 vertex values as B.C.
c     The membrane equation reads :
c        *******************************************
c        z(xi,eta) = a0 + a1*xi + a2*eta + a3*xi*eta
c        *******************************************
c     The partial derivatives, /dzx,dzy of /z are self-evident, but their
c     units are - for convenience - subject to "choice".

c     The SINGLE PRECISION data matrix (1x1 deg gridded) /zmat(180,360)
c     is available in common/datmat/. Effective lower left corner of the
c     patch has matrix indices: (i1,j1).
c     A   s p e c i a l   provision is made for data matrices in which
c     land is coded as a "negative" default value. A patch is not rejected
c     unless all four vertices are "land" (then the logical variable
c     pathok=.false., and results form this patch should not be used).
c     When some vertices are "land" then it still holds that pathok=.true.,
c     but the results are reported only on the "average" property of the
c     non-land vertices (the partial derivatives are set to zero).

c     Important convention on units:
c     x and y [deg]; if calling program uses [rad] for x and y , then
c             convert them to [deg] by division w/ torad before
c             calling /lapatch
c     z is always expressed in [u.u.]
c     dzx and dzy [u.u./choice] where "choice" is controlled by "mode":
c         if     (mode.eq.0) then "choice" = [deg],
c         elseif (mode.eq.1) then "choice" = [rad],
c         elseif (mode.eq.2) then "choice" = [km].

c     I n p u t :
c     ===========
c     x     =  D.P. longitude ( -imax,jmax )  [deg],
c     y     =  D.P. latitude  ( -imxh,imxh )  [deg],
c     mode  =  0  => [deg]  (denominator unit in partials, see above)
c           =  1  => [rad]
c           =  2  => [km]

c     O u t p u t :
c     ===========
c     z     =  D.P. dependent field variable [u.u],
c     dzx   =  D.P. partial derivative of /z wrt /x [u.u./"choice"],
c     dzy   =  D.P. partial derivative of /z wrt /y [u.u./"choice"],
c                   where "choice" is controlled by "mode".

c     ---------------------------------------------------------------
      implicit double precision(a-h,o-z)
      parameter( imax=180,imxh=imax/2,jmax=360,jmax2=2*jmax )
      parameter( dx=1.d0,dy=1.d0,quant=2*jmax+0.5d0 )
      parameter( torad=1.745329252d-02 )
      parameter( fay=111.195d0 )
      integer    kmz(4)
      real       zmat(imax,jmax),zhd(4),sum,zavg
      logical    pathok

      SAVE

  535 format(2f9.3,2i4)

c     if "patch" is "same" then i1=i1old, etc.,
c     then do not compute a new expansion
      common/same/    i1old,j1old
      common/datmat/  zmat
c     TEMPORARY : gradient of field
      common/zzz/ grad

      pathok = .true.

c     Assume matrix grid unit is "same" as unit of x and y : [degree]
c     longitude index of "lower left" ("South-West") corner of patch :
       itempbdd=ifix(real(x-0.5+jmax2))
      j1  =  mod(itempbdd,jmax) + 1
c      j1  =  mod(ifix(x-0.5+jmax2),jmax) + 1
c     latitude  index of "lower left" ("South-West") corner :
      i1  = min(imax,max(2,-y-0.5001d0+imxh+2))

      i2  =  i1
      j2  =  mod(ifix(real(x+0.5+jmax2)),jmax) + 1
c     (the above ensures the down-jump condition; e.g., from jmax to 1)
      i3  =  i1 - 1
      j3  =  j1
      i4  =  i3
      j4  =  j2

c     local co-ordinates: xi and eta on [0,1] :
      xi  =  x+quant - int(x+quant)
      eta =  y+quant - int(y+quant)

      zhd(1)= zmat(i1,j1)
      zhd(2)= zmat(i2,j2)
      zhd(3)= zmat(i3,j3)
      zhd(4)= zmat(i4,j4)

      if (i1.eq.i1old .and. j1.eq.j1old) goto 030

      nz    = 0
      mz    = 0
      sum   = 0.
      do 010 k=1,4
        if (zhd(k).ge.0.) then
          sum = sum + zhd(k)
          nz  = nz  + 1
        else
          mz  = mz  + 1
          kmz(mz) = k
        endif
  010 continue

      if (nz.lt.4) then
        if (nz.eq.0) then
cf          write( 0,*) ' path stopped by land, at i1,j1: ',i1,j1
cf          write( 0,*) ' RUN ALONG THIS PATH MUST NOT CONTINUE'
          pathok = .false.
          z   = zmat(i1,j1)
          dzx = 0.d0
          dzy = 0.d0
          return
c         and do not use any results from this patch !!
        else
          pathok = .true.
        endif
        zavg = sum / nz
        do 020 k=1,mz
          kz      = kmz(k)
          zhd(kz) = zavg
  020   continue
      endif

      a0    = zhd(1)
      a1    = zhd(2) -  a0
      a2    = zhd(3) -  a0
      a3    = zhd(4) - (a0+a1+a2)

  030 z     = a0 + a1*xi + a2*eta + a3*xi*eta
      if (nz.lt.4) then
        dzx = 0.d0
        dzy = 0.d0
      else
        dzx = a1 + a3*eta
        dzy = a2 + a3*xi
      endif

      if     (mode.eq.0) then
        return
      elseif (mode.eq.1) then
        dzx = dzx / torad
        dzy = dzy / torad
      elseif (mode.eq.2) then
        fax = fay*cos(y*torad)
        dzx = dzx / fax
        dzy = dzy / fay
      else
        write( 0,*)  ' mode bad ! ERROR: x,y,i1,j1 :'
        write( 0,535) x,y,i1,j1
        write(11,535) x,y,i1,j1
      endif
      return
      end
